@@ -7,7 +7,6 @@ except Exception:
     import socketserver  # py3
 
 import logging
-import traceback
 import json
 import base64
 import uuid
@@ -16,6 +15,7 @@ import importlib
 import socket
 import struct
 import time
+import traceback
 import weakref
 import functools
 import operator
@@ -42,6 +42,13 @@ except ImportError:
     # Nope, just normal python here
     EXCEPTION_TYPES = (Exception,)
 
+ENUM_TYPE = ()
+try:
+    from enum import Enum
+    ENUM_TYPE = (Enum,)
+except ImportError:  # py2 has no enum
+    pass
+
 
 class ThreadingTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
     # prevent server threads hanging around and stopping python from closing
@@ -56,6 +63,7 @@ MAX_VERSION = "max_v"
 MIN_VERSION = "min_v"
 COMMS_VERSION_1 = 1
 COMMS_VERSION_2 = 2
+COMMS_VERSION_3 = 3
 TYPE = "type"
 VALUE = "value"
 KEY = "key"
@@ -68,6 +76,7 @@ BOOL = "bool"
 STR = "str"
 BYTES = "bytes"
 NONE = "none"
+NOTIMPLEMENTED = "notimp"
 BRIDGED = "bridged"
 EXCEPTION = "exception"
 OBJ = "obj"
@@ -91,6 +100,7 @@ EVAL = "eval"
 EXPR = "expr"
 RESULT = "result"
 ERROR = "error"
+SHUTDOWN = "shutdown"
 
 HANDLE = "handle"
 NAME = "name"
@@ -100,18 +110,40 @@ KWARGS = "kwargs"
 
 BRIDGE_PREFIX = "_bridge"
 
-# Comms v2 (alpha) completely restructures the comms layer, breaking backwards compatability :(
-MIN_SUPPORTED_COMMS_VERSION = COMMS_VERSION_2
-MAX_SUPPORTED_COMMS_VERSION = COMMS_VERSION_2
+# Comms v3 (alpha) changes the bridged object representation to include the type - one day, I'll support backwards compatibility
+MIN_SUPPORTED_COMMS_VERSION = COMMS_VERSION_3
+MAX_SUPPORTED_COMMS_VERSION = COMMS_VERSION_3
 
 DEFAULT_RESPONSE_TIMEOUT = 2  # seconds
 
+GLOBAL_BRIDGE_SHUTDOWN = False
+
+# BridgedObjects have a little trouble with class methods (e.g., where the method of accessing is not instance.doThing(),  but more like
+# type(instance).doThing(instance) - such as __lt__, len(), str().
+# To handle this, we define a list of class methods that we want to expose - this is a little gross, I'd like to dynamically do this based on the methods in the
+# bridged object's type, but need to come up with a blacklist of things like __class__, __new__, etc which will interfere with the local objects first
+BRIDGED_CLASS_METHODS = ["__str__", "__len__", "__iter__"]
+# extract methods from operator, so I don't have to type out all the different options
+for operator_name in dir(operator):
+    # only do the methods that start and end with __, and exclude __new__
+    if operator_name.startswith("__") and operator_name.endswith("__") and operator_name != "__new__" and "builtin_function_or_method" in str(type(getattr(operator, operator_name))):
+        BRIDGED_CLASS_METHODS.append(operator_name)
+
 
 class BridgeException(Exception):
+    """ An exception happened on the other side of the bridge and has been proxied back here
+        The bridge is fine, but the remote code you ran might have had an issue.
+    """
+    pass
+
+
+class BridgeOperationException(Exception):
+    """ Some issue happened with the operation of the bridge itself. The bridge may not be in a good state """
     pass
 
 
 class BridgeClosedException(Exception):
+    """ The bridge has closed """
     pass
 
 
@@ -172,8 +204,6 @@ class BridgeCommandHandlerThread(threading.Thread):
     bridge_conn = None
     threadpool = None
 
-    ERROR_RESULT = json.dumps({ERROR: True})
-
     def __init__(self, threadpool):
         super(BridgeCommandHandlerThread, self).__init__()
 
@@ -190,12 +220,14 @@ class BridgeCommandHandlerThread(threading.Thread):
             while cmd is not None:  # get_command returns none if we should shut down
                 # handle a command and write back the response
                 # TODO make this return an error tied to the cmd_id, so it goes in the response mgr
-                result = BridgeCommandHandlerThread.ERROR_RESULT
+                result = None
                 try:
                     result = self.bridge_conn.handle_command(cmd)
                 except Exception as e:
                     self.bridge_conn.logger.error(
-                        "Unexpected exception for {}: {}".format(cmd, e))
+                        "Unexpected exception for {}: {}\n{}".format(cmd, e, traceback.format_exc()))
+                    # pack a minimal error, so the other end doesn't have to wait for a timeout
+                    result = json.dumps({VERSION: COMMS_VERSION_3, TYPE: ERROR, ID: cmd[ID], }).encode("utf-8")
 
                 try:
                     write_size_and_data_to_socket(
@@ -253,7 +285,7 @@ class BridgeCommandHandlerThreadPool(object):
         self.ready_threads.release()
 
         try:
-            while not self.shutdown_flag:
+            while not self.shutdown_flag and not GLOBAL_BRIDGE_SHUTDOWN:
                 # get the read lock, so we can see if there's anything to do
                 with self.command_list_read_lock:
                     if len(self.command_list) > 0:
@@ -294,7 +326,7 @@ class BridgeReceiverThread(threading.Thread):
         # threadpool to handle creating/running threads to handle commands
         threadpool = BridgeCommandHandlerThreadPool(self.bridge_conn)
 
-        while True:  # TODO shutdown flag
+        while not GLOBAL_BRIDGE_SHUTDOWN:
             try:
                 data = read_size_and_data_from_socket(
                     self.bridge_conn.get_socket())
@@ -309,8 +341,8 @@ class BridgeReceiverThread(threading.Thread):
                     "Recv loop received {}".format(msg_dict))
 
                 if can_handle_version(msg_dict):
-                    if msg_dict[TYPE] == RESULT:
-                        # handle a response
+                    if msg_dict[TYPE] in [RESULT, ERROR]:
+                        # handle a response or error
                         self.bridge_conn.response_mgr.add_response(msg_dict)
                     else:
                         # queue this and hand off to a worker threadpool
@@ -323,6 +355,8 @@ class BridgeReceiverThread(threading.Thread):
                 # eat exceptions and continue, don't want a bad message killing the recv loop
                 self.bridge_conn.logger.exception(e)
 
+        self.bridge_conn.logger.debug("Receiver thread shutdown")
+
 
 class BridgeCommandHandler(socketserver.BaseRequestHandler):
 
@@ -334,6 +368,10 @@ class BridgeCommandHandler(socketserver.BaseRequestHandler):
             # run the recv loop directly
             BridgeReceiverThread(BridgeConn(
                 self.server.bridge, self.request, response_timeout=self.server.bridge.response_timeout)).run()
+
+            # only get here if the client has requested we shutdown the bridge
+            self.server.bridge.logger.debug("Receiver thread exited - bridge shutdown requested")
+            self.server.bridge.shutdown()
         except BridgeClosedException:
             pass  # expected - the client has closed the connection
         except Exception as e:
@@ -352,7 +390,15 @@ class BridgeHandle(object):
         self.attrs = dir(local_obj)
 
     def to_dict(self):
-        return {HANDLE: self.handle, TYPE: type(self.local_obj).__name__, ATTRS: self.attrs, REPR: repr(self.local_obj)}
+        # extract the type name from the repr for the type
+        type_repr = repr(type(self.local_obj))
+        # expect it to be something like <class 'foo.bar'> or <type 'foo.bar'>
+        if "'" in type_repr:
+            type_name = type_repr.split("'")[1]
+        else:
+            # just use the repr straight up
+            type_name = type_repr
+        return {HANDLE: self.handle, TYPE: type_name, ATTRS: self.attrs, REPR: repr(self.local_obj)}
 
     def __str__(self):
         return "BridgeHandle({}: {})".format(self.handle, self.local_obj)
@@ -416,6 +462,11 @@ class BridgeResponseManager(object):
             raise Exception(
                 "Didn't receive response {} before timeout".format(response_id))
 
+        if TYPE in data:
+            if data[TYPE] == ERROR:
+                # problem with the bridge itself, raise an exception
+                raise BridgeOperationException(data)
+
         with self.response_lock:
             # delete the entry, we're done here
             del self.response_dict[response_id]
@@ -443,6 +494,11 @@ class BridgeConn(object):
         self.response_mgr = BridgeResponseManager()
         self.response_timeout = response_timeout
 
+        # keep a cache of types of objects we've created
+        # we'll keep all the types forever (including handles to bridgedcallables in them) because types are super-likely
+        # to be reused regularly, and we don't want to keep deleting them and then having to recreate them all the time.
+        self.cached_bridge_types = dict()
+
     def __del__(self):
         """ On teardown, make sure we close our socket to the remote bridge """
         with self.comms_lock:
@@ -478,7 +534,7 @@ class BridgeConn(object):
         # note: this needs to come before int, because apparently bools are instances of int (but not vice versa)
         if isinstance(data, bool):
             serialized_dict = {TYPE: BOOL, VALUE: str(data)}
-        elif isinstance(data, INTEGER_TYPES):
+        elif isinstance(data, INTEGER_TYPES) and not isinstance(data, ENUM_TYPE):  # don't treat py3 enums as ints - pass them as objects
             serialized_dict = {TYPE: INT, VALUE: str(data)}
         elif isinstance(data, float):
             serialized_dict = {TYPE: FLOAT, VALUE: str(data)}
@@ -509,6 +565,8 @@ class BridgeConn(object):
             serialized_dict = {TYPE: BRIDGED, VALUE: data._bridge_handle}
         elif isinstance(data, type(None)):
             serialized_dict = {TYPE: NONE}
+        elif isinstance(data, type(NotImplemented)):
+            serialized_dict = {TYPE: NOTIMPLEMENTED}
         else:
             # it's an object. assign a reference
             obj_type = CALLABLE_OBJ if callable(data) else OBJ
@@ -541,27 +599,15 @@ class BridgeConn(object):
 
             return result
         elif serial_dict[TYPE] == EXCEPTION:
-            raise BridgeException(self.deserialize_from_dict(serial_dict[MESSAGE]), BridgedObject(
-                self, serial_dict[VALUE]))
+            raise BridgeException(self.deserialize_from_dict(serial_dict[MESSAGE]), self.build_bridged_object(serial_dict[VALUE]))
         elif serial_dict[TYPE] == BRIDGED:
             return self.get_object_by_handle(serial_dict[VALUE])
         elif serial_dict[TYPE] == NONE:
             return None
+        elif serial_dict[TYPE] == NOTIMPLEMENTED:
+            return NotImplemented
         elif serial_dict[TYPE] == OBJ or serial_dict[TYPE] == CALLABLE_OBJ:
-            if serial_dict[TYPE] == CALLABLE_OBJ:
-                # note: assumes we're not going to get something that's iterable and callable at the same time (except types or Jython classes... which aren't actually iterable, they may just have __iter__)
-                assert "__iter__" not in serial_dict[VALUE][
-                    ATTRS] or serial_dict[VALUE][TYPE] in ["type", "Class"], "Found something callable and iterable at the same time: {}".format(serial_dict[VALUE][TYPE])
-                return BridgedCallable(self, serial_dict[VALUE])
-            elif "__iter__" in serial_dict[VALUE][ATTRS] and ("__next__" in serial_dict[VALUE][ATTRS] or "next" in serial_dict[VALUE][ATTRS]):
-                return BridgedIterableIterator(self, serial_dict[VALUE])
-            elif "__iter__" in serial_dict[VALUE][ATTRS]:
-                return BridgedIterable(self, serial_dict[VALUE])
-            elif "__next__" in serial_dict[VALUE][ATTRS] or "next" in serial_dict[VALUE][ATTRS]:
-                return BridgedIterator(self, serial_dict[VALUE])
-            else:
-                # just an object
-                return BridgedObject(self, serial_dict[VALUE])
+            return self.build_bridged_object(serial_dict[VALUE], callable=(serial_dict[TYPE] == CALLABLE_OBJ))
 
         raise Exception("Unhandled data {}".format(serial_dict))
 
@@ -584,7 +630,7 @@ class BridgeConn(object):
             If timeout override set, wait that many seconds, else wait for default response timeout
         """
         cmd_id = str(uuid.uuid4())  # used to link commands and responses
-        envelope_dict = {VERSION: COMMS_VERSION_2,
+        envelope_dict = {VERSION: COMMS_VERSION_3,
                          ID: cmd_id,
                          TYPE: CMD,
                          CMD: command_dict}
@@ -640,8 +686,17 @@ class BridgeConn(object):
         handle = args_dict[HANDLE]
         name = args_dict[NAME]
         value = self.deserialize_from_dict(args_dict[VALUE])
-        self.logger.debug(
-            "local_set: {}.{} = {}".format(handle, name, value))
+
+        if self.logger.getEffectiveLevel() <= logging.DEBUG:
+            try:
+                # we want to get log the deserialized values, because they're useful.
+                # but this also means a bad repr can break things. So we get ready to
+                # catch that and fallback to undeserialized values
+                self.logger.debug("local_set: {}.{} = {}".format(handle, name, value))
+            except Exception as e:
+                self.logger.debug("Failed to log deserialized arguments: {}\n{}".format(e, traceback.format_exc()))
+                self.logger.debug(
+                    "Falling back:\n\tlocal_set: {}.{} = {}".format(handle, name, args_dict[VALUE]))
 
         target = self.get_object_by_handle(handle)
         result = None
@@ -670,8 +725,18 @@ class BridgeConn(object):
         args = self.deserialize_from_dict(args_dict[ARGS])
         kwargs = self.deserialize_from_dict(args_dict[KWARGS])
 
-        self.logger.debug(
-            "local_call: {}({},{})".format(handle, args, kwargs))
+        if self.logger.getEffectiveLevel() <= logging.DEBUG:
+            try:
+                # we want to get log the deserialized values, because they're useful.
+                # but this also means a bad repr can break things. So we get ready to
+                # catch that and fallback to undeserialized values
+                self.logger.debug(
+                    "local_call: {}({},{})".format(handle, args, kwargs))
+            except Exception as e:
+                self.logger.debug("Failed to log deserialized arguments: {}\n{}".format(e, traceback.format_exc()))
+                self.logger.debug(
+                    "Falling back:\n\tlocal_call: {}({},{})".format(handle, args_dict[ARGS], args_dict[KWARGS]))
+
         result = None
         try:
             target_callable = self.get_object_by_handle(handle)
@@ -693,7 +758,12 @@ class BridgeConn(object):
     def remote_del(self, handle):
         self.logger.debug("remote_del {}".format(handle))
         command_dict = {CMD: DEL, ARGS: {HANDLE: handle}}
-        self.send_cmd(command_dict, get_response=False)
+        try:
+            self.send_cmd(command_dict, get_response=False)
+        except ConnectionError:
+            # get a lot of these when shutting down if the bridge connection has already been torn down before the bridged objects are deleted
+            # just ignore - we want to know if the other operations fail, but deleting failing we can probably get away with
+            pass
 
     def local_del(self, args_dict):
         handle = args_dict[HANDLE]
@@ -746,12 +816,21 @@ class BridgeConn(object):
         return self.deserialize_from_dict(self.send_cmd(command_dict))
 
     def local_create_type(self, args_dict):
-        name = str(args_dict[NAME]) # type name can't be unicode string in python2 - force to string
+        name = str(args_dict[NAME])  # type name can't be unicode string in python2 - force to string
         bases = self.deserialize_from_dict(args_dict[BASES])
         dct = self.deserialize_from_dict(args_dict[DICT])
 
-        self.logger.debug(
-            "local_create_type {}, {}, {}".format(name, bases, dct))
+        if self.logger.getEffectiveLevel() <= logging.DEBUG:
+            try:
+                # we want to get log the deserialized values, because they're useful.
+                # but this also means a bad repr can break things. So we get ready to
+                # catch that and fallback to undeserialized values
+                self.logger.debug("local_create_type {}, {}, {}".format(name, bases, dct))
+            except Exception as e:
+                self.logger.debug("Failed to log deserialized arguments: {}\n{}".format(e, traceback.format_exc()))
+                self.logger.debug(
+                    "Falling back:\n\tlocal_create_type {}, {}, {}".format(name, args_dict[BASES], args_dict[DICT]))
+
         result = None
 
         try:
@@ -803,8 +882,17 @@ class BridgeConn(object):
         test_object = args[OBJ]
         check_class_tuple = args[TUPLE]
 
-        self.logger.debug("local_isinstance({},{})".format(
-            test_object, check_class_tuple))
+        if self.logger.getEffectiveLevel() <= logging.DEBUG:
+            try:
+                # we want to get log the deserialized values, because they're useful.
+                # but this also means a bad repr can break things. So we get ready to
+                # catch that and fallback to undeserialized values
+                self.logger.debug("local_isinstance({},{})".format(
+                    test_object, check_class_tuple))
+            except Exception as e:
+                self.logger.debug("Failed to log deserialized arguments: {}\n{}".format(e, traceback.format_exc()))
+                self.logger.debug(
+                    "Falling back:\n\tlocal_isinstance({})".format(args_dict))
 
         # make sure every element is a local object on this side
         if _is_bridged_object(test_object):
@@ -836,8 +924,19 @@ class BridgeConn(object):
         args = self.deserialize_from_dict(args_dict)
 
         result = None
+
+        if self.logger.getEffectiveLevel() <= logging.DEBUG:
+            try:
+                # we want to get log the deserialized values, because they're useful.
+                # but this also means a bad repr can break things. So we get ready to
+                # catch that and fallback to undeserialized values
+                self.logger.debug("local_eval({},{})".format(args[EXPR], args[KWARGS]))
+            except Exception as e:
+                self.logger.debug("Failed to log deserialized arguments: {}\n{}".format(e, traceback.format_exc()))
+                self.logger.debug(
+                    "Falling back:\n\local_eval {}".format(args_dict))
+
         try:
-            self.logger.debug("local_eval({},{})".format(args[EXPR], args[KWARGS]))
             """ the import __main__ trick allows accessing all the variables that the bridge imports, 
             so evals will run within the global context of what started the bridge, and the arguments 
             supplied as kwargs will override that """
@@ -849,9 +948,27 @@ class BridgeConn(object):
 
         return self.serialize_to_dict(result)
 
-    def handle_command(self, message_dict):
+    def remote_shutdown(self):
+        self.logger.debug("remote_shutdown")
+        result = self.deserialize_from_dict(self.send_cmd({CMD: SHUTDOWN}))
+        print(result)
+        if SHUTDOWN in result and result[SHUTDOWN]:
+            # shutdown received - as a gross hack, send a followup that we don't expect to return, to unblock some loops and actually let things shutdown
+            self.send_cmd({CMD: SHUTDOWN}, get_response=False)
 
-        response_dict = {VERSION: COMMS_VERSION_2,
+        return result
+
+    def local_shutdown(self):
+        global GLOBAL_BRIDGE_SHUTDOWN
+
+        self.logger.debug("local_shutdown")
+
+        GLOBAL_BRIDGE_SHUTDOWN = True
+
+        return self.serialize_to_dict({SHUTDOWN: True})
+
+    def handle_command(self, message_dict):
+        response_dict = {VERSION: COMMS_VERSION_3,
                          ID: message_dict[ID],
                          TYPE: RESULT,
                          RESULT: {}}
@@ -878,10 +995,62 @@ class BridgeConn(object):
             response_dict[RESULT] = self.local_isinstance(command_dict[ARGS])
         elif command_dict[CMD] == EVAL:
             response_dict[RESULT] = self.local_eval(command_dict[ARGS])
+        elif command_dict[CMD] == SHUTDOWN:
+            response_dict[RESULT] = self.local_shutdown()
 
         self.logger.debug("Responding with {}".format(response_dict))
         return json.dumps(response_dict).encode("utf-8")
 
+    def get_bridge_type(self, bridged_obj_dict, callable=False):
+        # Get a dynamic bridging type from the cache based on the type name, or create it based on the type recovered from the instance bridge handle
+        bridge_handle = bridged_obj_dict[HANDLE]
+        type_name = bridged_obj_dict[TYPE]
+
+        # short circuit - any function-like thing, as well as any type (or java.lang.Class) becomes a BridgedCallable (need to invoke types/classes, so they're callable)
+        if type_name in ["type", "java.lang.Class", "function", "builtin_function_or_method", "instancemethod", "method_descriptor", "wrapper_descriptor"]:
+            return BridgedCallable
+        elif type_name == "module":
+            return BridgedObject
+
+        # if we've already handled this type, use the old one
+        if type_name in self.cached_bridge_types:
+            return self.cached_bridge_types[type_name]
+
+        self.logger.debug("Creating type " + type_name)
+        # need to create a type
+        # grab the remote type for the instance.
+        remote_type = self.remote_get_type(bridge_handle)
+
+        # create the class dict by getting any of the methods we're interested in
+        class_dict = {}
+        for method_name in BRIDGED_CLASS_METHODS:
+            if method_name in remote_type._bridge_attrs:
+                class_dict[method_name] = remote_type._bridged_get(method_name)
+
+        # handle a python2/3 compatibility issue - 3 uses truediv for /, 2 uses div unless you've imported
+        # __future__.division. Allow falling back to __div__ if __truediv__ requested but not present
+        if "__div__" in remote_type._bridge_attrs and "__truediv__" not in remote_type._bridge_attrs:
+            class_dict["__truediv__"] = remote_type._bridged_get("__div__")
+
+        # create the bases - any class level method which requires special implementation needs to add the relevant type
+        bases = (BridgedObject,)
+
+        if callable:
+            bases = (BridgedCallable, )
+        elif "__next__" in remote_type._bridge_attrs or "next" in remote_type._bridge_attrs:
+            bases = (BridgedIterator, )
+
+        local_type = type(str("_bridged_" + type_name), bases, class_dict)  # str to force it to non-unicode in py2
+        self.cached_bridge_types[type_name] = local_type
+
+        return local_type
+
+    def build_bridged_object(self, obj_dict, callable=False):
+        # construct a bridgedobject, including getting/creating a local dynamic type for its type
+        bridge_type = self.get_bridge_type(obj_dict, callable=callable)
+
+        return bridge_type(self, obj_dict)
+
 
 class BridgeServer(threading.Thread):
     """ Python2Python RPC bridge server 
@@ -929,9 +1098,10 @@ class BridgeServer(threading.Thread):
         self.shutdown()
 
     def shutdown(self):
-        self.logger.info("Shutting down bridge")
         if self.is_serving:
+            self.logger.info("Shutting down bridge")
             self.is_serving = False
+            self.server.shutdown()
             self.server.server_close()
 
 
@@ -970,7 +1140,8 @@ class BridgeClient(object):
         """
         return self.client.remote_eval(eval_string, timeout_override=timeout_override, **kwargs)
 
-    # TODO shutdown
+    def remote_shutdown(self):
+        return self.client.remote_shutdown()
 
 
 def _is_bridged_object(object):
@@ -1025,61 +1196,6 @@ def bridged_isinstance(test_object, class_or_tuple):
     return result
 
 
-# At the moment, BridgedObjects have trouble with class methods (e.g., where the method of accessing is not instance.doThing(),  but more like # type(instance).doThing(instance) - such as </__lt__, len(), str().
-# To handle this, we define a list of class methods that we want to expose, and add a BridgedClassMethod handler for each of them to BridgedObject.
-# This is kind of gross, because it means all the class methods are exposed even for bridged objects that don't have them
-# TODO: possibly a better fix is to dynamically create BridgedObject subclasses on the fly for each different type of bridged object, with only the
-# appropriate class methods defined.
-BRIDGED_CLASS_METHODS = ["__str__", "__len__"]
-# extract methods from operator, so I don't have to type out all the different options
-for operator_name in dir(operator):
-    # only do the methods that start and end with __, and exclude __new__
-    if operator_name.startswith("__") and operator_name.endswith("__") and operator_name != "__new__" and "builtin_function_or_method" in str(type(getattr(operator, operator_name))):
-        BRIDGED_CLASS_METHODS.append(operator_name)
-
-# We define a class for the BridgedClassMethod so we can use descriptor get to resolve which method should be used for a given instance
-
-
-class BridgedClassMethod(object):
-    method_name = None
-
-    def __init__(self, method_name):
-        self.method_name = method_name
-
-    def __get__(self, instance, owner):
-        """ Use descriptor get so that we bind the BridgedClassMethod to correctly to the remote method and the instance when it's grabbed.
-            Use functools.partial to return a wrapper to the remote method with the instance object as the first arg
-        """
-        remote_type = instance._bridged_get_type()
-
-        remote_method = None
-
-        # so, jython getattr() will query first the object, then its parent class for a java type, this
-        # will end up querying the jython type, then java.lang.Class
-        # java.lang.Class does have the rich comparables (__lt__/__gt__/etc) defined, but they appear to
-        # expect to be called differently (will return an error saying they expected 2 args, got 2 args)
-        # so what we'll do instead is check if the class methods we're looking for appear in the attributes
-        # of the type before we try to get them - if not, raise an exception.
-        # This does lead to a difference in behaviour for ghidra_bridge: currentProgram < currentProgram is a
-        # TypeError/not supported on the bridge end, and False in the ghidra python interpreter. However,
-        # since the java.lang.Class level implementation doesn't seem particularly useful, I don't think
-        # that loses too much.
-        # TODO is there a cleaner way to handle this? Can we push this logic into _bridged_get, or will
-        # it break things there?
-        if self.method_name in remote_type._bridge_attrs:
-            remote_method = remote_type._bridged_get(self.method_name)
-        elif (self.method_name == "__truediv__") and ("__div__" in remote_type._bridge_attrs):
-            # handle a python2/3 compatibility issue - 3 uses truediv for /, 2 uses div unless you've imported
-            # __future__.division. Allow falling back to __div__ if __truediv__ requested but not present
-            # TODO pull that out into its own handling class?
-            remote_method = remote_type._bridged_get("__div__")
-
-        if remote_method is None:
-            raise AttributeError()
-
-        return functools.partial(remote_method, instance)
-
-
 class BridgedObject(object):
     """ An object you can only interact with on the opposite side of a bridge """
     _bridge_conn = None
@@ -1091,7 +1207,7 @@ class BridgedObject(object):
 
     # list of methods which we don't bridge, but need to have specific names (so we can't use the _bridge prefix for them)
     # TODO decorator to mark a function as local, don't bridge it - then have it automatically fill this out (also needs to work for subclasses)
-    _LOCAL_METHODS = ["__del__", "__str__", "__repr__", "__dir__", "getdoc"]
+    _LOCAL_METHODS = ["__del__", "__str__", "__repr__", "__dir__", "__bool__", "__nonzero__", "getdoc"]
 
     # list of attrs that we don't want to waste bridge calls on
     _DONT_BRIDGE = ["__mro_entries__",  # ignore mro entries - only being called if we're creating a class based off a bridged object
@@ -1130,7 +1246,7 @@ class BridgedObject(object):
                 result = self._bridged_get(attr)
             except BridgeException as be:
                 # unwrap AttributeErrors if they occurred on the other side of the bridge
-                if be.args[1]._bridge_type == "AttributeError":
+                if be.args[1]._bridge_type.endswith("AttributeError"):
                     raise AttributeError(be.args[0])
                 else:
                     # some other cause - just reraise the exception
@@ -1207,10 +1323,11 @@ class BridgedObject(object):
     def __dir__(self):
         return dir(super(type(self))) + (self._bridge_attrs if self._bridge_attrs else [])
 
+    def __bool__(self):
+        # py3 vs 2 - __bool__ vs __nonzero__
+        return self._bridge_conn.remote_eval("bool(x)", x=self)
 
-# after BridgedObject is defined, update it to include handlers for common class methods
-for method_name in BRIDGED_CLASS_METHODS:
-    setattr(BridgedObject, method_name, BridgedClassMethod(method_name))
+    __nonzero__ = __bool__  # handle being run in a py2 environment
 
 
 class BridgedCallable(BridgedObject):
@@ -1256,13 +1373,6 @@ class BridgedCallable(BridgedObject):
         """
         return functools.partial(self, instance)
 
-# TODO remove this
-
-
-class BridgedIterable(BridgedObject):
-    def __iter__(self):
-        return self._bridged_get("__iter__")()
-
 
 class BridgedIterator(BridgedObject):
     def __next__(self):
@@ -1271,14 +1381,9 @@ class BridgedIterator(BridgedObject):
             return self._bridged_get("__next__" if "__next__" in self._bridge_attrs else "next")()
         except BridgeException as e:
             # we expect the StopIteration exception - check to see if that's what we got, and if so, raise locally
-            if e.args[1]._bridge_type == "StopIteration":
+            if e.args[1]._bridge_type.endswith("StopIteration"):
                 raise StopIteration
             # otherwise, something went bad - reraise
             raise
 
     next = __next__  # handle being run in a py2 environment
-
-
-class BridgedIterableIterator(BridgedIterator, BridgedIterable):
-    """ Common enough that iterables return themselves from __iter__ """
-    pass
